@@ -1,7 +1,7 @@
 /* SPDX-License-Identifier: BSD-2-Clause */
 /*
  * Linux interface driver for dhcpcd
- * Copyright (c) 2006-2023 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2024 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -102,13 +102,6 @@ int if_getssid_wext(const char *ifname, uint8_t *ssid);
 #define IFLA_WIRELESS (IFLA_MASTER + 1)
 #endif
 
-/* For some reason, glibc doesn't include newer flags from linux/if.h
- * However, we cannot include linux/if.h directly as it conflicts
- * with the glibc version. D'oh! */
-#ifndef IFF_LOWER_UP
-#define IFF_LOWER_UP	0x10000		/* driver signals L1 up		*/
-#endif
-
 /* Buggy CentOS and RedHat */
 #ifndef SOL_NETLINK
 #define	SOL_NETLINK	270
@@ -542,6 +535,12 @@ static int if_carrier_from_flags(unsigned int flags)
 	return ((flags & (IFF_LOWER_UP | IFF_RUNNING)) ==
 		(IFF_LOWER_UP | IFF_RUNNING))
 #ifdef IFF_DORMANT
+		/*
+		 * IFF_DORMANT means L1 is up but waiting for an external
+		 * event, for example 802.1X
+		 * We treat this as DOWN, but then return true for if_roaming()
+		 * so that the interface status is persisted.
+		 */
 		&& !(flags & IFF_DORMANT)
 #endif
 		? LINK_UP : LINK_DOWN;
@@ -556,21 +555,27 @@ if_carrier(struct interface *ifp, __unused const void *ifadata)
 	return if_carrier_from_flags(ifp->flags);
 }
 
+
 bool
 if_roaming(struct interface *ifp)
 {
 
-	if (!ifp->wireless ||
-	    ifp->flags & IFF_RUNNING)
-		return false;
-
-#if defined(IFF_DORMANT)
-	return ifp->flags & IFF_DORMANT;
-#elif defined(IFF_LOWER_UP)
-	return (ifp->flags & (IFF_UP|IFF_LOWER_UP)) == (IFF_UP|IFF_LOWER_UP);
-#else
-	return false;
+	return
+#ifdef IFF_DORMANT
+	   ifp->flags & IFF_DORMANT ||
 #endif
+#ifdef IFF_LOWER_UP
+	   /*
+	    * IFF_DORMANT only occurs for supplicant initiated roaming.
+	    * For firmware initiated roaming we don't get IFF_DORMANT.
+	    * Seems weird that the driver can't set it though.
+	    * We can check that IFF_RUNNING is not set but UP and L1 are
+	    * to get the same effect.
+	    */
+	   (ifp->flags & (IFF_UP | IFF_LOWER_UP | IFF_RUNNING)) ==
+	       (IFF_UP | IFF_LOWER_UP) ||
+#endif
+	    false;
 }
 
 int
@@ -804,7 +809,7 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 	int ret;
 #endif
 #ifdef INET6
-	struct in6_addr addr6;
+	struct in6_addr *local6 = NULL, *address6 = NULL;
 	int flags;
 #endif
 
@@ -882,19 +887,24 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 #endif
 #ifdef INET6
 	case AF_INET6:
-		memset(&addr6, 0, sizeof(addr6));
 		for (; RTA_OK(rta, len); rta = RTA_NEXT(rta, len)) {
-			switch (rta->rta_type) {
+		switch (rta->rta_type) {
 			case IFA_ADDRESS:
-				memcpy(&addr6.s6_addr, RTA_DATA(rta),
-				       sizeof(addr6.s6_addr));
+				address6 = (struct in6_addr *)RTA_DATA(rta);
+				break;
+			case IFA_LOCAL:
+				local6 = (struct in6_addr *)RTA_DATA(rta);
 				break;
 			}
 		}
+		if (local6 != NULL)
+			address6 = local6;
+		if (address6 == NULL)
+			break; /* should be impossible */
 
 		/* Validate RTM_DELADDR really means address deleted
 		 * and anything else really means address exists. */
-		flags = if_addrflags6(ifp, &addr6, NULL);
+		flags = if_addrflags6(ifp, address6, NULL);
 		if (nlm->nlmsg_type == RTM_DELADDR) {
 			if (flags != -1)
 				break;
@@ -904,7 +914,7 @@ link_addr(struct dhcpcd_ctx *ctx, struct interface *ifp, struct nlmsghdr *nlm)
 		}
 
 		ipv6_handleifa(ctx, nlm->nlmsg_type, NULL, ifp->name,
-		    &addr6, ifa->ifa_prefixlen, ifa->ifa_flags,
+		    address6, ifa->ifa_prefixlen, ifa->ifa_flags,
 		    (pid_t)nlm->nlmsg_pid);
 		break;
 #endif
